import tensorflow as tf
import math

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'


class NeuralNetworkGraph:

    def __init__(self, input_size, layer_sizes, output_size):
        """
        Args:
          input_size: number of input channels
          layer_sizes: Sizes of hidden layers in a tuple or list.
          output_size: Number of output channels.
        """
        self.input_size = int(input_size)
        self.output_size = int(output_size)
        self.layer_sizes = self._set_layer_sizes(layer_sizes)
        self.num_hidden_layers = len(self.layer_sizes)-1
        self.layers = []
        self.prediction_op = None

    def _set_layer_sizes(self, layer_sizes):
        layer_sizes = tuple(filter(None, layer_sizes))
        if layer_sizes[-1] < self.output_size:
            raise ValueError('Last layer size must be greater or equal output size')
        return (self.input_size,) + layer_sizes

    def build_neural_network(self, input):
        """Builds a neural network with the given layers and output size.

        Args:
          input: Images placeholder, from inputs().

        Returns:
          logits: Output tensor with the computed logits.
        """

        assert isinstance(input, tf.Tensor)
        assert self.input_size == int(input.get_shape()[1])
        assert self.layers == []

        self.input_placeholder = input
        # self.batch_size = int(input.get_shape()[0])

        self.layers.append(input)
        for i in range(1, self.num_hidden_layers+1):
            self.layers.append(
                self._add_layer(
                    'layer%04d' % i, self.layer_sizes[i-1], self.layer_sizes[i], self.layers[i-1], tf.nn.relu
                )
            )

        logits = self._add_layer('output', self.layer_sizes[-1], self.output_size, self.layers[-1])

        self.layers.append(logits)

        return logits

    def loss(self, logits, labels):
        """Calculates the loss from the logits and the labels.

        Args:
          logits: Logits tensor, float - [batch_size, NUM_CLASSES].
          labels: Labels tensor, int32 - [batch_size].

        Returns:
          loss: Loss tensor of type float.
        """
        onehot_labels = self._convert_labels_to_onehot(labels)
        cross_entropy = tf.nn.softmax_cross_entropy_with_logits(
            logits, onehot_labels, name='cross_entropy'
        )
        return tf.reduce_mean(cross_entropy, name='cross_entropy_mean')

    def training(self, loss, learning_rate):
        """Sets up the training Ops.

        Creates a summarizer to track the loss over time in TensorBoard.

        Creates an optimizer and applies the gradients to all trainable variables.

        The Op returned by this function is what must be passed to the
        `sess.run()` call to cause the model to train.

        Args:
          loss: Loss tensor, from loss().
          learning_rate: The learning rate to use for gradient descent.

        Returns:
          train_op: The Op for training.
      """
        # Add a scalar summary for the snapshot loss.
        tf.scalar_summary(loss.op.name, loss)
        # Create the gradient descent optimizer with the given learning rate.
        optimizer = tf.train.GradientDescentOptimizer(learning_rate)
        # Create a variable to track the global step.
        global_step = tf.Variable(0, name='global_step', trainable=False)
        # Use the optimizer to apply the gradients that minimize the loss
        # (and also increment the global step counter) as a single training step.
        train_op = optimizer.minimize(loss, global_step=global_step)
        return train_op

    def evaluation(self, logits, labels):
        """Evaluate the quality of the logits at predicting the label.

        Args:
          logits: Logits tensor, float - [batch_size, NUM_CLASSES].
          labels: Labels tensor, int32 - [batch_size], with values in the
            range [0, NUM_CLASSES).

        Returns:
          A scalar int32 tensor with the number of examples (out of batch_size)
          that were predicted correctly.
        """
        # For a classifier model, we can use the in_top_k Op. It returns a bool
        # tensor with shape [batch_size] that is true for the examples where the
        # label's is was in the top k (here k=1) of all logits for that example.
        correct = tf.nn.in_top_k(logits, labels, 1)
        # Return the number of true entries.
        return tf.reduce_sum(tf.cast(correct, tf.int32))

    def predict(self, session, image):
        if self.prediction_op is None:
            self.prediction_op = tf.argmax(self.layers[-1], 1)

        image_data = image.reshape(self.input_size)
        feed_dict = {self.input_placeholder: [image_data]}
        best = session.run(self.prediction_op, feed_dict)
        return best[0]


    def _add_layer(self, layer_name, in_units_size, out_units_size, input_layer, function=lambda x: x):
        with tf.name_scope(layer_name):
            weights = self._initialize_weights(in_units_size, out_units_size)
            biases = self._initialize_biases(out_units_size)
            new_layer = function(tf.matmul(input_layer, weights) + biases)
        return new_layer

    def _initialize_weights(self, in_units_size, out_units_size):
        """initialize weights with small amount of noise for symmetry breaking, and to prevent 0 gradients"""
        return tf.Variable(
            tf.truncated_normal([in_units_size, out_units_size], stddev=1.0 / math.sqrt(float(in_units_size))),
            name='weights'
        )

    def _initialize_biases(self, out_units_size):
        return tf.Variable(tf.zeros([out_units_size]), name='biases')

    def _convert_labels_to_onehot(self, labels):
        """
        Convert from sparse integer labels in the range [0, NUM_CLASSSES) to 1-hot dense float
        vectors (that is we will have batch_size vectors, each with NUM_CLASSES values, all of
        which are 0.0 except there will be a 1.0 in the entry corresponding to the label).
        """
        batch_size = tf.size(labels)
        labels = tf.expand_dims(labels, 1)
        indices = tf.expand_dims(tf.range(0, batch_size), 1)
        concated = tf.concat(1, [indices, labels])
        return tf.sparse_to_dense(concated, tf.pack([batch_size, self.output_size]), 1.0, 0.0)




from nn_wtf.images_labels_data_set import ImagesLabelsDataSet
from .util import create_minimal_input_placeholder

import numpy

import unittest

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'

NUM_TRAINING_SAMPLES = 20
IMAGE_WIDTH = 10
IMAGE_HEIGHT = 10
BATCH_SIZE = 5

class ImagesLabelsDataSetTest(unittest.TestCase):

    def test_init_without_fake_data_runs(self):
        _create_empty_data_set()

    def test_init_with_fake_data_runs(self):
        images = create_minimal_input_placeholder()
        ImagesLabelsDataSet(images, images, fake_data=True)

    def test_init_with_different_label_size_fails(self):
        images = create_empty_image_data()
        labels = create_empty_label_dataof_size(NUM_TRAINING_SAMPLES+1)
        with self.assertRaises(AssertionError):
            ImagesLabelsDataSet(images, labels)

    def test_init_with_wrong_label_shape_fails(self):
        images = create_empty_image_data()
        with self.assertRaises(AssertionError):
            ImagesLabelsDataSet(images, images)

    def test_next_batch_returns_correct_data_format(self):
        data_set = _create_empty_data_set()
        images, labels = data_set.next_batch(BATCH_SIZE)
        self.assertIsInstance(images, numpy.ndarray)
        self.assertEqual(2, len(images.shape))
        self.assertEqual(BATCH_SIZE, images.shape[0])
        self.assertEqual(IMAGE_WIDTH*IMAGE_HEIGHT, images.shape[1])
        self.assertIsInstance(labels, numpy.ndarray)
        self.assertEqual(1, len(labels.shape))
        self.assertEqual(BATCH_SIZE, labels.shape[0])

    def test_next_batch_runs_repeatedly(self):
        data_set = _create_empty_data_set()
        batch_size = NUM_TRAINING_SAMPLES//2
        _, _ = data_set.next_batch(batch_size)
        _, _ = data_set.next_batch(batch_size)

    def test_next_batch_sets_epochs_completed(self):
        data_set = _create_empty_data_set()
        batch_size = NUM_TRAINING_SAMPLES//2
        self.assertEqual(0, data_set.epochs_completed)
        _, _ = data_set.next_batch(batch_size)
        self.assertEqual(0, data_set.epochs_completed)
        _, _ = data_set.next_batch(batch_size)
        self.assertEqual(0, data_set.epochs_completed)
        _, _ = data_set.next_batch(batch_size)
        self.assertEqual(1, data_set.epochs_completed)


def _create_empty_data_set():
    images = create_empty_image_data()
    labels = create_empty_label_data()
    return ImagesLabelsDataSet(images, labels)


def create_empty_image_data():
    buf = [0] * NUM_TRAINING_SAMPLES * IMAGE_WIDTH * IMAGE_HEIGHT
    data = numpy.fromiter(buf, dtype=numpy.uint8)
    return data.reshape(NUM_TRAINING_SAMPLES, IMAGE_WIDTH, IMAGE_HEIGHT, 1)


def create_empty_label_data():
    return create_empty_label_dataof_size(NUM_TRAINING_SAMPLES)

def create_empty_label_dataof_size(size):
    buf = [0] * size
    return numpy.fromiter(buf, dtype=numpy.uint8)



import tensorflow as tf

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'

MINIMAL_INPUT_SIZE = 2
MINIMAL_LAYER_GEOMETRY = (2, 2)
MINIMAL_OUTPUT_SIZE = 2
MINIMAL_BATCH_SIZE = 2


def create_minimal_input_placeholder():
    return tf.placeholder(tf.float32, shape=(MINIMAL_BATCH_SIZE, MINIMAL_INPUT_SIZE))


def get_project_root_folder():
    import os
    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Functions for downloading and reading MNIST data."""

import gzip
import os

import numpy

from nn_wtf.images_labels_data_set import ImagesLabelsDataSet
from nn_wtf.mnist_graph import IMAGE_SIZE

import urllib.request

SOURCE_URL = 'http://yann.lecun.com/exdb/mnist/'


def maybe_download(filename, work_directory):
  """Download the data from Yann's website, unless it's already here."""
  if not os.path.exists(work_directory):
    os.mkdir(work_directory)
  filepath = os.path.join(work_directory, filename)
  if not os.path.exists(filepath):
    filepath, _ = urllib.request.urlretrieve(SOURCE_URL + filename, filepath)
    statinfo = os.stat(filepath)
    print('Successfully downloaded', filename, statinfo.st_size, 'bytes.')
  return filepath


def _read32(bytestream):
  dt = numpy.dtype(numpy.uint32).newbyteorder('>')
  return numpy.frombuffer(bytestream.read(4), dtype=dt)[0]


def extract_images(filename):
  """Extract the images into a 4D uint8 numpy array [index, y, x, depth]."""
  print('Extracting', filename)
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2051:
      raise ValueError(
          'Invalid magic number %d in MNIST image file: %s' %
          (magic, filename))
    num_images = _read32(bytestream)
    rows = _read32(bytestream)
    cols = _read32(bytestream)
    return images_from_bytestream(bytestream, rows, cols, num_images)


def images_from_bytestream(bytestream, rows, cols, num_images):
    buf = bytestream.read(rows * cols * num_images)
    data = numpy.frombuffer(buf, dtype=numpy.uint8)
    return data.reshape(num_images, rows, cols, 1)


def read_one_image_from_file(filename):
    with open(filename, 'rb') as bytestream:
        return read_one_image_from_bytestream(bytestream)


def read_one_image_from_url(url):
    with urllib.request.urlopen(url) as bytestream:
        return read_one_image_from_bytestream(bytestream)


def read_one_image_from_bytestream(bytestream):
    rows, cols = IMAGE_SIZE, IMAGE_SIZE
    return images_from_bytestream(bytestream, rows, cols, 1)


def dense_to_one_hot(labels_dense, num_classes=10):
    """Convert class labels from scalars to one-hot vectors."""
    num_labels = labels_dense.shape[0]
    index_offset = numpy.arange(num_labels) * num_classes
    labels_one_hot = numpy.zeros((num_labels, num_classes))
    labels_one_hot.flat[index_offset + labels_dense.ravel()] = 1
    return labels_one_hot


def extract_labels(filename, one_hot=False):
  """Extract the labels into a 1D uint8 numpy array [index]."""
  print('Extracting', filename)
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2049:
      raise ValueError(
          'Invalid magic number %d in MNIST label file: %s' %
          (magic, filename))
    num_items = _read32(bytestream)
    buf = bytestream.read(num_items)
    labels = numpy.frombuffer(buf, dtype=numpy.uint8)
    if one_hot:
      return dense_to_one_hot(labels)
    return labels


class DataSets:

    def __init__(self, train, validation, test):
        self.train = test
        self.validation = validation
        self.test = test

TRAIN_IMAGES = 'train-images-idx3-ubyte.gz'
TRAIN_LABELS = 'train-labels-idx1-ubyte.gz'
TEST_IMAGES = 't10k-images-idx3-ubyte.gz'
TEST_LABELS = 't10k-labels-idx1-ubyte.gz'
VALIDATION_SIZE = 5000


def read_data_sets(train_dir, one_hot=False):

    local_file = maybe_download(TRAIN_IMAGES, train_dir)
    train_images = extract_images(local_file)

    local_file = maybe_download(TRAIN_LABELS, train_dir)
    train_labels = extract_labels(local_file, one_hot=one_hot)

    local_file = maybe_download(TEST_IMAGES, train_dir)
    test_images = extract_images(local_file)

    local_file = maybe_download(TEST_LABELS, train_dir)
    test_labels = extract_labels(local_file, one_hot=one_hot)

    validation_images = train_images[:VALIDATION_SIZE]
    validation_labels = train_labels[:VALIDATION_SIZE]
    train_images = train_images[VALIDATION_SIZE:]
    train_labels = train_labels[VALIDATION_SIZE:]

    return DataSets(
        ImagesLabelsDataSet(train_images, train_labels),
        ImagesLabelsDataSet(validation_images, validation_labels),
        ImagesLabelsDataSet(test_images, test_labels)
    )


def fake_data_sets(one_hot):
    return DataSets(
        ImagesLabelsDataSet([], [], fake_data=True, one_hot=one_hot),
        ImagesLabelsDataSet([], [], fake_data=True, one_hot=one_hot),
        ImagesLabelsDataSet([], [], fake_data=True, one_hot=one_hot)
    )



from nn_wtf.images_labels_data_set import ImagesLabelsDataSet
from nn_wtf.input_data import read_one_image_from_file

from nn_wtf.mnist_graph import IMAGE_SIZE

from .util import get_project_root_folder

import numpy

import unittest

__author__ = 'Lene Preuss <lene.preuss@gmail.com>'


class InputDataTest(unittest.TestCase):

    def setUp(self):
        self.data = read_one_image_from_file(get_project_root_folder()+'/nn_wtf/data/0.raw')

    def test_read_one_image_from_file(self):
        self.assertIsInstance(self.data, numpy.ndarray)
        self.assertEqual(4, len(self.data.shape))
        self.assertEqual(1, self.data.shape[0])
        self.assertEqual(IMAGE_SIZE, self.data.shape[1])
        self.assertEqual(IMAGE_SIZE, self.data.shape[2])
        self.assertEqual(1, self.data.shape[3])

    def test_image_labels_data_set_from_image(self):
        labels = numpy.fromiter([0], dtype=numpy.uint8)
        data_set = ImagesLabelsDataSet(self.data, labels)




from nn_wtf.tests.neural_network_graph_test import NeuralNetworkGraphTest
from nn_wtf.tests.images_labels_data_set_test import ImagesLabelsDataSetTest
from nn_wtf.tests.input_data_test import InputDataTest

import unittest

__author__ = 'Lene Preuss <lene.preuss@gmx.net>'

unittest.main()



import numpy

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'


class ImagesLabelsDataSet:

    def __init__(self, images, labels, fake_data=False, one_hot=False):
        """Construct a DataSet. one_hot arg is used only if fake_data is true.

        Args:
          images: 4D numpy.ndarray of shape (num images, image height, image width, image depth)
          labels: 1D numpy.ndarray of shape (num images)
        """

        if fake_data:
            self._num_examples = 10000
            self.one_hot = one_hot
        else:
            _check_constructor_arguments_valid(images, labels)

            self._num_examples = images.shape[0]

            # Convert shape from [num examples, rows, columns, depth] to [num examples, rows*columns]
            # TODO: assumes depth == 1
            images = images.reshape(images.shape[0], images.shape[1] * images.shape[2])
            images = normalize(images)

        self._images = images
        self._labels = labels
        self._epochs_completed = 0
        self._index_in_epoch = 0

    @property
    def images(self):
        return self._images

    @property
    def labels(self):
        return self._labels

    @property
    def num_examples(self):
        return self._num_examples

    @property
    def epochs_completed(self):
        return self._epochs_completed

    def next_batch(self, batch_size, fake_data=False):
        """Return the next `batch_size` examples from this data set."""
        if fake_data:
            return self._fake_batch(batch_size)

        return self._next_batch_in_epoch(batch_size)

    def _next_batch_in_epoch(self, batch_size):
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        if self._index_in_epoch > self._num_examples:
            # Finished epoch
            self._epochs_completed += 1
            # Shuffle the data
            self._shuffle_data()
            # Start next epoch
            start = 0
            self._index_in_epoch = batch_size
            assert batch_size <= self._num_examples
        end = self._index_in_epoch
        return self._images[start:end], self._labels[start:end]

    def _shuffle_data(self):
        perm = numpy.arange(self._num_examples)
        numpy.random.shuffle(perm)
        self._images = self._images[perm]
        self._labels = self._labels[perm]

    def _fake_batch(self, batch_size):
      fake_image = [1] * 784
      if self.one_hot:
        fake_label = [1] + [0] * 9
      else:
        fake_label = 0
      return [fake_image for _ in range(batch_size)], [fake_label for _ in range(batch_size)]


def normalize(ndarray):
    """Transform a ndarray that contains uint8 values to floats between 0. and 1.

    :param ndarray:
    :return:
    """
    ndarray = ndarray.astype(numpy.float32)
    return numpy.multiply(ndarray, 1.0 / 255.0)

def invert(ndarray):
    return numpy.subtract(1.0, ndarray)

def _check_constructor_arguments_valid(images, labels):
    assert isinstance(images, numpy.ndarray), \
        'images not of type numpy.ndarray, but ' + type(images).__name__
    assert isinstance(labels, numpy.ndarray), \
        'labels not of type numpy.ndarray, but ' + type(images).__name__
    assert len(images.shape) == 4, \
        'images must have 4 dimensions: number of images, image height, image width, color depth'
    assert len(labels.shape) == 1, 'labels must have one dimension: number of labels'
    assert images.shape[0] == labels.shape[0], \
        'number of images: {}, number of labels: {}'.format(images.shape[0], labels.shape[0])
    assert images.shape[3] == 1, 'image depth must be 1'




import time
import pprint
import tensorflow as tf

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'


class NeuralNetworkOptimizer:

    DEFAULT_LEARNING_RATE = 0.1
    DEFAULT_LAYER_SIZES = (
        (32, 48, 64, 80, 96, 128),
        (32, 48, 64, 80, 96, 128),
        (None, 16, 32, 48)
    )

    class TimingInfo:

        def __init__(self, cpu_time, wall_time, precision, step, layers):
            self.cpu_time = cpu_time
            self.wall_time = wall_time
            self.precision = precision
            self.step = step
            self.layers = layers

        def __str__(self):
            return 'CPU: {:7.3f}s Wall: {:7.3f}s Precision: {:5.2f}% Iterations: {:4d} Geometry: {}'.format(
                self.cpu_time, self.wall_time, 100.*self.precision, self.step, str(self.layers)
            )

        def __repr__(self):
            return str(self.__dict__())

        def __dict__(self):
            return {'cpu_time': self.cpu_time, 'step': self.step, 'layers': self.layers}

    def __init__(self, tested_network, training_precision, layer_sizes=None, learning_rate=None, verbose=False):
        self.tested_network = tested_network
        self.verbose = verbose
        self.learning_rate = learning_rate if learning_rate else self.DEFAULT_LEARNING_RATE
        self.training_precision = training_precision
        self.layer_sizes = self.DEFAULT_LAYER_SIZES if layer_sizes is None else layer_sizes

    def brute_force_optimal_network_geometry(self, data_sets, max_steps):
        results = self.time_all_tested_geometries(data_sets, max_steps)
        return results[0].layers

    def time_all_tested_geometries(self, data_sets, max_steps):
        results = []
        for layer1_size, layer2_size, layer3_size in self.get_network_geometries():
            run_info = self.timed_run_training(
                data_sets, layer1_size, layer2_size, layer3_size, max_steps=max_steps
            )
            if self.verbose: print(run_info)
            results.append(run_info)
        results = sorted(results, key=lambda r: r.cpu_time)
        if self.verbose: pprint.pprint(results, width=100)
        return results

    def get_network_geometries(self):
        return ((l1, l2, l3)
                for l1 in self.layer_sizes[0]
                for l2 in self.layer_sizes[1] if l2 <= l1
                for l3 in self.layer_sizes[2] if l3 is None or l3 <= l2)

    def brute_force_optimize_learning_rate(self):
        raise NotImplemented()

    def timed_run_training(self, data_sets, layer1_size, layer2_size, layer3_size, max_steps=10000):
        graph, cpu, wall = timed_run(self.run_training_once, data_sets, layer1_size, layer2_size, layer3_size, max_steps)
        return self.TimingInfo(cpu, wall, graph.precision, graph.step, (layer1_size, layer2_size, layer3_size))

    def run_training_once(self, data_sets, layer1_size, layer2_size, layer3_size, max_steps):
        # Tell TensorFlow that the model will be built into the default Graph.
        with tf.Graph().as_default():
            graph = self.tested_network(
                learning_rate=self.learning_rate,
                hidden1=layer1_size, hidden2=layer2_size, hidden3=layer3_size,
                verbose=False
            )
            graph.train(data_sets, max_steps, precision=self.training_precision, steps_between_checks=50)
        return graph


def timed_run(function, *args, **kwargs):
        start_cpu_time, start_wall_time = time.process_time(), time.time()
        returned = function(*args, **kwargs)
        return returned, time.process_time()-start_cpu_time, time.time()-start_wall_time



from pprint import pprint

from nn_wtf.neural_network_graph import NeuralNetworkGraph

from .util import MINIMAL_INPUT_SIZE, MINIMAL_OUTPUT_SIZE, MINIMAL_LAYER_GEOMETRY, MINIMAL_BATCH_SIZE
from .util import create_minimal_input_placeholder

import tensorflow as tf

import unittest

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'



class NeuralNetworkGraphTest(unittest.TestCase):

    def test_init_runs(self):
        NeuralNetworkGraph(MINIMAL_INPUT_SIZE, MINIMAL_LAYER_GEOMETRY, MINIMAL_OUTPUT_SIZE)

    def test_init_fails_on_bad_layer_sizes(self):
        with self.assertRaises(TypeError):
            NeuralNetworkGraph(2, 2, 2)

    def test_init_fails_if_last_layer_smaller_than_output_size(self):
        with self.assertRaises(ValueError):
            NeuralNetworkGraph(2, (2, 1), 2)

    def test_build_neural_network_runs(self):
        graph = self._create_minimal_graph()
        input_placeholder = create_minimal_input_placeholder()

        graph.build_neural_network(input_placeholder)

    def test_build_neural_network_runs_only_once(self):
        graph = self._create_minimal_graph()
        input_placeholder = create_minimal_input_placeholder()

        graph.build_neural_network(input_placeholder)

        with self.assertRaises(AssertionError):
            graph.build_neural_network(input_placeholder)

    def test_build_neural_network_fails_on_bad_input_size(self):
        graph = self._create_minimal_graph()
        input_placeholder = tf.placeholder(tf.float32, shape=(MINIMAL_BATCH_SIZE, MINIMAL_INPUT_SIZE+1))

        with self.assertRaises(AssertionError):
            graph.build_neural_network(input_placeholder)

    def test_build_neural_network_output(self):
        graph = self._create_minimal_graph()

        output = graph.build_neural_network(create_minimal_input_placeholder())

        self.assertIsInstance(output, tf.Tensor)
        self.assertEqual(2, output.get_shape().ndims)
        self.assertEqual(MINIMAL_OUTPUT_SIZE, int(output.get_shape()[1]))
        self.assertEqual(len(graph.layers)-2, graph.num_hidden_layers)
        self.assertEqual(len(MINIMAL_LAYER_GEOMETRY), graph.num_hidden_layers)

    def test_build_neural_network_output_with_three_layers(self):
        self._check_num_hidden_layers_for_input_is((4, 3, 2), 3)

    def test_build_neural_network_output_with_last_layer_none(self):
        self._check_num_hidden_layers_for_input_is((4, 3, None), 2)

    def test_build_neural_network_output_with_middle_layer_none(self):
        self._check_num_hidden_layers_for_input_is((4, None, 2), 2)

    def _check_num_hidden_layers_for_input_is(self, definition, expected_size):
        graph = NeuralNetworkGraph(MINIMAL_INPUT_SIZE, definition, MINIMAL_OUTPUT_SIZE)
        graph.build_neural_network(create_minimal_input_placeholder())
        self.assertEqual(expected_size, graph.num_hidden_layers)

    def _create_minimal_graph(self):
        return NeuralNetworkGraph(MINIMAL_INPUT_SIZE, MINIMAL_LAYER_GEOMETRY, MINIMAL_OUTPUT_SIZE)





from setuptools import setup

try:
    from pypandoc import convert

    def read_md(f):
        return convert(f, 'rst')
except ImportError:
    def read_md(f):
        with open(f, 'r') as h:
            return h.read()


setup(
    name='nn_wtf',
    version='0.1.5',
    description='Neural Networks Wrapper for TensorFlow',
    long_description=read_md('README.md'),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='machine learning neural networks tensorflow',
    url='http://github.com/lene/nn-wtf',
    author='Lene Preuss',
    author_email='lene.preuss@gmail.com',
    license='Apache Software License V2',
    packages=['nn_wtf'],
    package_data={'nn_wtf': ['data/*.raw']},
    install_requires=[
        'numpy'
    ],
    dependency_links=[
        'https://storage.googleapis.com/tensorflow/linux/cpu/tensorflow-0.7.0-py3-none-linux_x86_64.whl'
    ],
    include_package_data=True,
    zip_safe=False
)



import time
import math

import tensorflow as tf

from nn_wtf.neural_network_graph import NeuralNetworkGraph

__author__ = 'Lene Preuss <lp@sinnwerkstatt.com>'


# The MNIST dataset has 10 classes, representing the digits 0 through 9.
NUM_CLASSES = 10

# The MNIST images are always 28x28 pixels.
IMAGE_SIZE = 28
IMAGE_PIXELS = IMAGE_SIZE * IMAGE_SIZE

DEFAULT_TRAIN_DIR='.nn_wtf-data'


class MNISTGraph:

    def __init__(
        self, verbose=True,
        learning_rate=0.01, hidden1=128, hidden2=32, hidden3=None, batch_size=100,
        train_dir=DEFAULT_TRAIN_DIR
    ):
        self.verbose = verbose
        self.learning_rate = learning_rate
        self.hidden = (hidden1, hidden2)
        if hidden3:
            self.hidden += (hidden3,)
        self.batch_size = batch_size
        self.train_dir = ensure_is_dir(train_dir)
        self.fake_data = False

        self._build_graph()
        self._setup_summaries()

    def train(self, data_sets, max_steps, precision=None, steps_between_checks=100):

        self.session = self.initialize_session()

        # And then after everything is built, start the training loop.
        for self.step in range(max_steps):
            start_time = time.time()

            # Fill a feed dictionary with the actual set of images and labels for this particular
            # training step.
            feed_dict = self.fill_feed_dict(data_sets.train)

            # Run one step of the model.  The return values are the activations from the `train_op`
            # (which is discarded) and the `loss` Op. To inspect the values of your Ops or
            # variables, you may include them in the list passed to session.run() and the value
            # tensors will be returned in the tuple from the call.
            _, loss_value = self.session.run([self.train_op, self.loss], feed_dict=feed_dict)

            duration = time.time() - start_time

            # Write the summaries and print an overview fairly often.
            if self.step % steps_between_checks == 0:
                self.write_summary(duration, feed_dict, loss_value, self.step)
                if precision is not None:
                    self.do_eval(data_sets.test)
                    if self.precision > precision:
                        return

            # Save a checkpoint and evaluate the model periodically.
            if (self.step + 1) % 1000 == 0 or (self.step + 1) == max_steps:
                self.saver.save(self.session, save_path=self.train_dir, global_step=self.step)
                self.print_evaluations(data_sets)

    def print_evaluations(self, data_sets):
        if self.verbose: print('Training Data Eval:')
        self.print_eval(data_sets.train)

        if self.verbose: print('Validation Data Eval:')
        self.print_eval(data_sets.validation)

        if self.verbose: print('Test Data Eval:')
        self.print_eval(data_sets.test)

    def initialize_session(self):
        # Create a session for running Ops on the Graph.
        sess = tf.Session()
        # Run the Op to initialize the variables.
        init = tf.initialize_all_variables()
        sess.run(init)
        # Instantiate a SummaryWriter to output summaries and the Graph.
        self.summary_writer = tf.train.SummaryWriter(self.train_dir, graph_def=sess.graph_def)
        return sess

    def fill_feed_dict(self, data_set):
        """Fills the feed_dict for training the given step.

        A feed_dict takes the form of:
        feed_dict = {
            <placeholder>: <tensor of values to be passed for placeholder>,
              ....
        }

        Args:
          data_set: The set of images and labels, from input_data.read_data_sets()

        Returns:
          feed_dict: The feed dictionary mapping from placeholders to values.
        """
        # Create the feed_dict for the placeholders filled with the next `batch size ` examples.
        images_feed, labels_feed = data_set.next_batch(self.batch_size, self.fake_data)
        feed_dict = {
            self.images_placeholder: images_feed,
            self.labels_placeholder: labels_feed,
        }
        return feed_dict

    def write_summary(self, duration, feed_dict, loss_value, step):
        # Print status to stdout.
        if self.verbose:
            print('Step %d: loss = %.2f (%.3f sec)' % (step, loss_value, duration))
        # Update the events file.
        summary_str = self.session.run(self.summary_op, feed_dict=feed_dict)
        self.summary_writer.add_summary(summary_str, step)

    def do_eval(self, data_set):
        """Runs one evaluation against the full epoch of data.

        Args:
          session: The session in which the model has been trained.
          data_set: The set of images and labels to evaluate, from
            input_data.read_data_sets().
        """
        self.true_count = 0  # Counts the number of correct predictions.
        steps_per_epoch = data_set.num_examples // self.batch_size
        self.num_examples = steps_per_epoch * self.batch_size
        for _ in range(steps_per_epoch):
            feed_dict = self.fill_feed_dict(data_set)
            self.true_count += self.session.run(self.eval_correct, feed_dict=feed_dict)
        self.precision = self.true_count / self.num_examples

    def print_eval(self, data_set):
        self.do_eval(data_set)
        if self.verbose:
            print('  Num examples: %d  Num correct: %d  Precision @ 1: %0.04f' %
                  (self.num_examples, self.true_count, self.precision))

    def evaluate_new_data_set(self, data_set):
        self.batch_size = data_set.num_examples
        self.num_examples = data_set.num_examples
        self.print_eval(data_set)

    def predict(self, image):
        return self.graph.predict(self.session, image)

    def _build_graph(self):
        # Generate placeholders for the images and labels.
        self.images_placeholder, self.labels_placeholder = placeholder_inputs(self.batch_size)

        self.graph = NeuralNetworkGraph(
            self.images_placeholder.get_shape()[1], self.hidden, NUM_CLASSES
        )

        # Build a Graph that computes predictions from the inference model.
        self.logits = self.graph.build_neural_network(
            self.images_placeholder
        )

        # Add to the Graph the Ops for loss calculation.
        self.loss = self.graph.loss(self.logits, self.labels_placeholder)

        # Add to the Graph the Ops that calculate and apply gradients.
        self.train_op = self.graph.training(self.loss, self.learning_rate)

        # Add the Op to compare the logits to the labels during evaluation.
        self.eval_correct = self.graph.evaluation(self.logits, self.labels_placeholder)

    def _setup_summaries(self):
        # Build the summary operation based on the TF collection of Summaries.
        self.summary_op = tf.merge_all_summaries()
        # Create a saver for writing training checkpoints.
        self.saver = tf.train.Saver()
        self.summary_writer = None


def ensure_is_dir(train_dir_string):
    if not train_dir_string[-1] == '/':
        train_dir_string += '/'
    return train_dir_string


def placeholder_inputs(batch_size):
    """Generate placeholder variables to represent the input tensors.

    These placeholders are used as inputs by the rest of the model building
    code and will be fed from the downloaded data in the .run() loop, below.

    Args:
      batch_size: The batch size will be baked into both placeholders.

    Returns:
      images_placeholder: Images placeholder.
      labels_placeholder: Labels placeholder.
    """
    # Note that the shapes of the placeholders match the shapes of the full image and label
    # tensors, except the first dimension is now batch_size rather than the full size of
    # the train or test data sets.
    # images_placeholder = tf.placeholder(tf.float32, shape=(batch_size, IMAGE_PIXELS))
    images_placeholder = tf.placeholder(tf.float32, shape=(None, IMAGE_PIXELS), name='images')
    # labels_placeholder = tf.placeholder(tf.int32, shape=batch_size)
    labels_placeholder = tf.placeholder(tf.int32, shape=(None,), name='labels')
    return images_placeholder, labels_placeholder

